/*
 * Created on 16-okt-2003
 *
*/
package persistence;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;

import org.w3c.dom.Node;

import persistence.exceptions.ClassNotRegisteredException;
import persistence.exceptions.CompTypeRequiredException;
import persistence.exceptions.MissingContentTagException;
import persistence.script.DBScript;
import persistence.table.TableComponent;
import persistence.visitor.CreateVisitor;
import persistence.visitor.DeleteVisitor;
import persistence.visitor.DropVisitor;
import persistence.visitor.InsertVisitor;
import persistence.visitor.QueryCompTypeVisitor;
import persistence.visitor.QueryJoinVisitor;
import persistence.visitor.QueryRegisterVisitor;
import persistence.visitor.ReadVisitor;
import persistence.visitor.TableComponentVisitor;
import persistence.visitor.UpdateVisitor;

/**
 * Implementatie van AbstractSQLFactory.
 * @author bram
 *
 * 16-okt-2003
 */
public class AbstractSQLFactoryImpl implements AbstractSQLFactory {
	//Contains (tablename,[table-attributesmap,extendstype,contained-boolean])	 	 
	private TableComponent persistenceTree;
	private String fullPrincipalClassName;
	private String connectionExpression;
	private static DBScript script = new DBScript(AbstractSQLFactory.registry);

	static {
		//in finalize() van Script wordt een create-/drop-script gemaakt voor ALLE behandelde
		//klassen
		System.runFinalizersOnExit(true);
	}

	/**
	 * coll contains all classes which will have to be persisted. This is 
	 * needed for inheritance checks.
	 * @param collection
	 */
	public static void setAllClassesToPersist(Collection coll) {
		PDLParser.setAllClassesToPersist(coll);
	}

	public AbstractSQLFactoryImpl(
		Node involvedNode,
		String fullPrincipalClassName,
		String connectionExpression)
		throws CompTypeRequiredException, MissingContentTagException {
		this.connectionExpression = connectionExpression;
		setCurrentNodeList(involvedNode, fullPrincipalClassName);
	}

	/**
	 * Sets a NodeList of &lt;persistentClass&gt;-elements as the current one and extraction of useful info happens.
	 * @param node
	 * @param fullPrincipalClassName
	 */
	public void setCurrentNodeList(Node node, String fullPrincipalClassName)
		throws CompTypeRequiredException, MissingContentTagException {
		this.fullPrincipalClassName = fullPrincipalClassName;
		PDLParser pdlParser = new PDLParser(node, this.fullPrincipalClassName);
		persistenceTree = pdlParser.parse();
		AbstractSQLFactory.registry.addPersistenceTree(
			this.fullPrincipalClassName,
			persistenceTree);
	}

	/**
	 * @see persistence.AbstractSQLFactory#getInsertStatement(String)
	 */
	public String getInsertStatement() {
		TableComponentVisitor vis =
			new InsertVisitor(fullPrincipalClassName, connectionExpression);
		persistenceTree.accept(vis);
		return vis.getResult();
	}

	/**
	 * @see persistence.AbstractSQLFactory#getReadStatement(String)
	 */
	public String getReadStatement() {
		TableComponentVisitor vis =
			new ReadVisitor(fullPrincipalClassName, connectionExpression);
		persistenceTree.accept(vis);
		return vis.getResult();
	}

	/* (non-Javadoc)
	 * @see persistence.AbstractSQLFactory#getUpdateStatement(org.w3c.dom.NodeList)
	 */
	public String getUpdateStatement() {
		TableComponentVisitor vis =
			new UpdateVisitor(fullPrincipalClassName, connectionExpression);
		persistenceTree.accept(vis);
		return vis.getResult();
	}

	/* (non-Javadoc)
	 * @see persistence.AbstractSQLFactory#getDeleteStatement(org.w3c.dom.NodeList)
	 */
	public String getDeleteStatement() {
		TableComponentVisitor vis =
			new DeleteVisitor(fullPrincipalClassName, connectionExpression);
		persistenceTree.accept(vis);
		return vis.getResult();
	}

	/**
	 * One has to establish a Connection named conn, and to catch Exceptions.
	 * @see persistence.AbstractSQLFactory#getCreateStatement(String)
	 */
	public String getCreateStatement(String dialectVariableName) {
		String res = "try{\n" + "conn.setAutoCommit(false);\n";
		TableComponentVisitor vis =
			new CreateVisitor("conn", dialectVariableName);
		persistenceTree.accept(vis);
		res += vis.getResult();

		res =
			res
				+ "conn.setAutoCommit(true);\n"
				+ "}catch(Exception e){\n"
				+ "try{\n"
				+ "System.out.println(\"Create-problem near "
				+ fullPrincipalClassName
				+ "!\");\n"
				+ "e.printStackTrace();\n"
				+ "conn.rollback();\n"
				+ "conn.setAutoCommit(true);\n"
				+ "}catch(SQLException s){\n"
				+ "s.printStackTrace();\n"
				+ "}\n"
				+ "}\n";
		return res;
	}

	/**
	 * One has to establish a Connection named conn, and to catch Exceptions.
	 * @see persistence.AbstractSQLFactory#getDropStatement(String)
	 */
	public String getDropStatement(String dialectVariableName) {
		String res = "try{\n" + "conn.setAutoCommit(false);\n";
		TableComponentVisitor vis =
			new DropVisitor("conn", dialectVariableName);
		persistenceTree.accept(vis);
		res += vis.getResult();

		//preparedStatement gebruiken bij arrays en collections en maps!!
		res =
			res
				+ "conn.setAutoCommit(true);\n"
				+ "}catch(Exception e){\n"
				+ "try{\n"
				+ "System.out.println(\"Drop-problem near "
				+ fullPrincipalClassName
				+ "!\");\n"
				+ "e.printStackTrace();\n"
				+ "conn.rollback();\n"
				+ "conn.setAutoCommit(true);\n"
				+ "}catch(SQLException s){\n"
				+ "s.printStackTrace();\n"
				+ "}\n"
				+ "}\n";
		return res;
	}

	public static void createScript() {
		script.generate();
	}

	public void addToScript() {
		script.addCreate(getCreateStatement("_d_i_a_l_e_c_t_"));
		script.addDrop(getDropStatement("_d_i_a_l_e_c_t_"));
	}

	private String getQueryJoinStatement() {
		TableComponentVisitor vis =
			new QueryJoinVisitor(fullPrincipalClassName);
		persistenceTree.accept(vis);
		return vis.getResult();
	}

	private String getQueryRegisterStatement() {
		TableComponentVisitor vis =
			new QueryRegisterVisitor(fullPrincipalClassName);
		persistenceTree.accept(vis);
		return vis.getResult();
	}

	private String getQueryCompTypeStatement() {
		TableComponentVisitor vis =
			new QueryCompTypeVisitor(fullPrincipalClassName);
		persistenceTree.accept(vis);
		return vis.getResult();
	}

	/**
	 * @see persistence.AbstractSQLFactory#getQueryStatements()
	 */
	public String getQueryStatements() {
		return getQueryJoinStatement()
			+ "\n"
			+ getQueryRegisterStatement()
			+ "\n"
			+ getQueryCompTypeStatement();
	}

}
